package com.uas.platform.oos.core.oos.client;

import com.amazonaws.services.s3.AmazonS3;
import com.amazonaws.services.s3.model.ObjectMetadata;
import com.amazonaws.services.s3.model.S3Object;
import com.amazonaws.services.s3.model.S3ObjectInputStream;
import com.uas.platform.oos.core.oos.AbstractClient;
import com.uas.platform.oos.core.oos.FileOperations;
import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.util.Map;
import org.apache.commons.io.FilenameUtils;
import org.apache.commons.lang3.RandomStringUtils;
import org.apache.commons.lang3.Validate;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.util.StringUtils;
import org.springframework.web.multipart.MultipartFile;

/**
 * A helper class provides OOS objects operations.
 *
 * @author huxz
 */
public class OosClientHelper extends AbstractClient implements FileOperations {

  private static final Log logger = LogFactory.getLog(OosClientHelper.class);

  /**
   * A bucket name used to test.
   */
  private static final String BUCKET_FOR_TEST = "uas.test";

  private static final String GROUP_TAG_NAME = "group";

  @Override
  public String upload(String bucketName, File file) {
    Validate.notNull(file, "Uploading file can't be Null");
    if (StringUtils.isEmpty(bucketName)) {
      bucketName = BUCKET_FOR_TEST;
    }

    AmazonS3 client = getClient();

    // Check whether bucket exists
    if (!client.doesBucketExist(bucketName)) {
      throw new IllegalArgumentException(String.format("Bucket {%s} doesn't exist", bucketName));
    }

    String key = generateObjectKey(file.getName());
    client.putObject(bucketName, key, file);

    return spliceResourceUrl(bucketName, key);
  }

  @Override
  public String upload(String bucketName, MultipartFile multiFile) {
    Validate.notNull(multiFile, "Uploading file can't be Null");
    if (StringUtils.isEmpty(bucketName)) {
      bucketName = BUCKET_FOR_TEST;
    }

    AmazonS3 client = getClient();

    // Check whether bucket exists
    if (!client.doesBucketExist(bucketName)) {
      throw new IllegalArgumentException(String.format("Bucket {%s} doesn't exist", bucketName));
    }


    final String key = generateObjectKey(multiFile.getOriginalFilename());

    ObjectMetadata metadata = new ObjectMetadata();
    metadata.setContentLength(multiFile.getSize());
    metadata.setContentType(multiFile.getContentType());

    try {
      client.putObject(bucketName, key, multiFile.getInputStream(), metadata);

      return spliceResourceUrl(bucketName, key);
    } catch (IOException e) {
      throw new IllegalStateException("Read file failed", e);
    }
  }

  @Override
  public String upload(String bucketName, String url) {
    Validate.notBlank(url, "Remote file url must be non-empty");
    if (StringUtils.isEmpty(bucketName)) {
      bucketName = BUCKET_FOR_TEST;
    }

    AmazonS3 client = getClient();

    // Check whether bucket exists
    if (!client.doesBucketExist(bucketName)) {
      throw new IllegalArgumentException(String.format("Bucket {%s} doesn't exist", bucketName));
    }

    try {
      URL resource = new URL(url);
      HttpURLConnection connection = (HttpURLConnection) resource.openConnection();
      connection.setReadTimeout(30 * 1000);
      connection.connect();

      String extension = FilenameUtils.getExtension(url);

      // Cache remote file to local before upload to OOS
      File tempFile = File.createTempFile("oos-remote", "." + extension);
      FileOutputStream fos = new FileOutputStream(tempFile);
      BufferedInputStream bis = new BufferedInputStream(connection.getInputStream());

      byte[] buf = new byte[1024];
      int len;
      while ((len = bis.read(buf)) >= 0) {
        fos.write(buf, 0 ,len);
      }
      fos.close();
      bis.close();

      String filePath = upload(bucketName, tempFile);
      //noinspection ResultOfMethodCallIgnored
      tempFile.delete();

      return filePath;
    } catch (IOException e) {
      throw new IllegalStateException("Read file failed", e);
    }
  }

  @Override
  public S3ObjectInputStream download(String url) {
    validateUrl(url);

    FileInfo info = new FileInfo(url).getFileInfo();

    S3Object s3Object = getClient().getObject(info.bucketName, info.key);
    if (s3Object == null) {
      throw new IllegalStateException("File doesn't exist");
    }
    return s3Object.getObjectContent();
  }

  @Override
  public void delete(String url) {
    validateUrl(url);

    FileInfo info = new FileInfo(url).getFileInfo();

    getClient().deleteObject(info.bucketName, info.key);
  }

  @Override
  public Map<String, Object> showMetadata(String url) {
    validateUrl(url);

    FileInfo info = new FileInfo(url).getFileInfo();

    ObjectMetadata metadata = getClient().getObjectMetadata(info.bucketName, info.key);
    if (metadata == null) {
      throw new IllegalStateException("File doesn't exist");
    }
    return metadata.getRawMetadata();
  }

  /**
   * Validate url which is used to download, delete or get metadata.
   *
   * @param url link of remote file
   */
  private void validateUrl(String url) {
    Validate.notBlank(url, "File url must be non-empty");
    if (!url.contains(OOS_ENDPOINT)) {
      throw new IllegalArgumentException("This url doesn't belong to OOS");
    }
  }

  /**
   * Generate object key for files to be uploaded.
   *
   * @param filename  the name of file uploading
   * @return  object key generated by random string
   */
  private String generateObjectKey(String filename) {
    final String extension = FilenameUtils.getExtension(filename);

    String[] randomStrs = new String[4];
    randomStrs[0] = GROUP_TAG_NAME + RandomStringUtils.randomNumeric(1);
    randomStrs[1] = RandomStringUtils.randomAlphabetic(1).toUpperCase() +
        RandomStringUtils.randomNumeric(2);
    randomStrs[2] = RandomStringUtils.randomAlphanumeric(2);
    randomStrs[3] = RandomStringUtils.randomAlphanumeric(2);
    final String prefix = String.join("/", randomStrs);

    final String baseName = "CgpkyFk" + RandomStringUtils.randomAlphanumeric(24);

    return prefix + "/" + baseName + "." + extension;
  }

  /**
   * Gain resource url when file upload successfully.
   *
   * @param bucketName  the name of bucket
   * @param key the key of Object
   * @return  resource url
   */
  private String spliceResourceUrl(String bucketName, String key) {
    return "http://" + bucketName + "." +
        OOS_ENDPOINT + "/" +
        key;
  }

  /**
   * Method class gains bucket and key from url.
   */
  private class FileInfo {

    private String url;
    private String bucketName;
    private String key;

    FileInfo(String url) {
      this.url = url;
    }

    FileInfo getFileInfo() {
      String pathUrl = url.replaceFirst("http://", "")
          .replaceFirst("." + OOS_ENDPOINT, "");
      int index = pathUrl.indexOf("/");
      bucketName = pathUrl.substring(0, index);
      key = pathUrl.substring(index + 1);

      if (logger.isDebugEnabled()) {
        logger.debug("BucketName: " + bucketName + ", Key: " + key);
      }
      return this;
    }
  }
}
