package com.uas.platform.oos.dashboard.api;

import com.uas.platform.oos.core.common.ResultBean;
import com.uas.platform.oos.dashboard.service.FileOperationService;
import java.io.BufferedInputStream;
import java.io.IOException;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpHeaders;
import org.springframework.http.HttpStatus;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.util.FileCopyUtils;
import org.springframework.web.bind.annotation.DeleteMapping;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.PostMapping;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestParam;
import org.springframework.web.bind.annotation.RestController;
import org.springframework.web.multipart.MultipartFile;

/**
 * Http api interface of file operations.
 *
 * @author huxz
 */
@RestController
@RequestMapping(value = "/file")
public class FileOperationController {

  private final FileOperationService operationService;

  @Autowired
  public FileOperationController(
      FileOperationService operationService) {
    this.operationService = operationService;
  }

  /**
   * Upload file to OOS.
   *
   * @param file  an uploaded file
   * @param bucketName  bucket name
   * @return  http://<Bucket-Name>.oos-gz.ctyunapi.cn/groupX/XXX/XX/XX/XX...XX.jpg
   */
  @PostMapping(value = "/upload", produces = MediaType.APPLICATION_JSON_UTF8_VALUE)
  public ResultBean<String> uploadFile(@RequestParam(name = "file") MultipartFile file,
      @RequestParam(defaultValue = "uas.test") String bucketName) {
    return new ResultBean<>(operationService.uploadFileToRemote(bucketName, file));
  }

  /**
   * Upload remote file to OOS.
   *
   * @param url an url of remote file
   * @param bucketName  bucket name
   * @return  http://<Bucket-Name>.oos-gz.ctyunapi.cn/groupX/XXX/XX/XX/XX...XX.jpg
   */
  @PostMapping(value = "/upload-remote", produces = MediaType.APPLICATION_JSON_UTF8_VALUE)
  public ResultBean<String> uploadRemoteFile(@RequestParam String url,
      @RequestParam(defaultValue = "uas.test") String bucketName) {
    return new ResultBean<>(operationService.uploadRemoteFileToOos(bucketName, url));
  }

  /**
   * Download an uploaded file.
   *
   * @param url an url of uploaded file
   * @return  file output stream
   */
  @GetMapping(value = "/download")
  public ResponseEntity<byte[]> downloadFile(@RequestParam String url) throws IOException {

    try (BufferedInputStream bis = new BufferedInputStream(operationService.download(url))) {
      String fileName = url.substring(url.lastIndexOf("/") + 1);

      HttpHeaders headers = new HttpHeaders();
      headers.setContentDispositionFormData("attachment", fileName);
      headers.setContentType(MediaType.APPLICATION_OCTET_STREAM);

      // Copy input stream to byte array, and create response entity
      return new ResponseEntity<>(FileCopyUtils.copyToByteArray(bis), headers, HttpStatus.CREATED);
    }
  }

  /**
   * Delete an uploaded file from OOS.
   *
   * @param url an url of uploaded file
   * @return  true, if operation succeed
   */
  @DeleteMapping(value = "/delete", produces = MediaType.APPLICATION_JSON_UTF8_VALUE)
  public ResultBean<Boolean> removeFile(@RequestParam String url) {
    return new ResultBean<>(operationService.delete(url));
  }
}
